<?php

namespace Drupal\Tests\commerce_datatrans\Kernel;

use Drupal\commerce_order\Entity\OrderItemType;
use Drupal\commerce_payment\Entity\Payment;
use Drupal\commerce_payment\Entity\PaymentGateway;
use Drupal\commerce_price\Price;
use Drupal\Tests\commerce\Kernel\CommerceKernelTestBase;
use Drupal\commerce_datatrans\Plugin\Commerce\PaymentGateway\Datatrans;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Psr7\Response;
use Symfony\Component\HttpFoundation\Request;

/**
 * @coversDefaultClass \Drupal\commerce_datatrans\Plugin\Commerce\PaymentGateway\Datatrans
 *
 * @group commerce_datatrans
 */
class DatatransKernelTest extends CommerceKernelTestBase {

  /**
   * The entity type manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * The order.
   *
   * @var \Drupal\commerce_order\Entity\OrderInterface
   */
  protected $order;

  /**
   * The payment gateway config entity.
   *
   * @var \Drupal\commerce_payment\Entity\PaymentGatewayInterface
   */
  protected $paymentGateway;

  public static $modules = [
    'entity_reference_revisions',
    'state_machine',
    'profile',
    'commerce_number_pattern',
    'commerce_order',
    'commerce_payment',
    'commerce_datatrans',
  ];

  /**
   * {@inheritdoc}
   */
  protected function setUp() {
    parent::setUp();

    $this->installEntitySchema('commerce_order');
    $this->installEntitySchema('commerce_order_item');
    $this->installEntitySchema('commerce_payment');
    $this->installConfig(['commerce_order']);

    OrderItemType::create([
      'id' => 'default',
      'label' => 'Default',
      'orderType' => 'default',
    ])->save();

    $this->entityTypeManager = $this->container->get('entity_type.manager');

    // Create an order.
    $this->order = $this->entityTypeManager->getStorage('commerce_order')->create([
      'type' => 'default',
      'order_number' => '1',
      'store_id' => $this->store->id(),
      'state' => 'draft',
    ]);
    $this->order->save();

    $order_item = $this->entityTypeManager->getStorage('commerce_order_item')->create([
      'type' => 'default',
      'unit_price' => [
        'number' => '999',
        'currency_code' => 'USD',
      ],
    ]);
    $order_item->save();
    $this->order->setItems([$order_item]);
    $this->order->save();

    $this->paymentGateway = PaymentGateway::create([
      'id' => 'datatrans_id',
      'plugin' => 'datatrans',
    ]);
    $this->paymentGateway->save();
  }

  /**
   * @covers ::onReturn
   * @covers ::validateResponseData
   * @covers ::processPayment
   */
  public function testOnReturnComplete() {
    $datatrans = $this->getDatatransGateway();

    // Create the request and call onReturn().
    $request = Request::create('', 'POST');
    $request->request->add([
      'refno' => $this->order->id(),
      'status' => 'success',
      'security_level' => 2,
      'uppTransactionId' => 123,
      'authorizationCode' => 456,
      'amount' => 999,
      'currency' => 'USD',
      'sign2' => 'f6ecaff66282456e3a949e0ef19daf4453bf70682bda2e981fd1193ea9da7833',
      'responseMessage' => 'test',
      'reqtype' => 'CAA',
    ]);
    $datatrans->onReturn($this->order, $request);

    // Process a second time, ensure that no duplicate is created.
    $datatrans->onReturn($this->order, $request);

    $payments = $this->entityTypeManager->getStorage('commerce_payment')->loadMultiple(NULL);
    $this->assertCount(1, $payments);

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = reset($payments);

    $this->assertEquals('completed', $payment->getState()->value);
    $this->assertEquals(123, $payment->getRemoteId());
    $this->assertEquals($this->order->getTotalPrice(), $payment->getAmount());
    $this->assertEquals('datatrans_id', $payment->getPaymentGatewayId());
  }

  /**
   * @covers ::onReturn
   * @covers ::validateResponseData
   * @covers ::processPayment
   */
  public function testOnReturnAuthorizeOnly() {
    // Initialize the datatrans gateway.
    $datatrans = $this->getDatatransGateway();

    // Create the request and call onReturn().
    $request = Request::create('', 'POST');
    $request->request->add([
      'refno' => $this->order->id(),
      'status' => 'success',
      'security_level' => 2,
      'uppTransactionId' => 123,
      'authorizationCode' => 456,
      'amount' => 999,
      'currency' => 'USD',
      'sign2' => 'f6ecaff66282456e3a949e0ef19daf4453bf70682bda2e981fd1193ea9da7833',
      'responseMessage' => 'test',
      'reqtype' => 'NOA',
    ]);
    $datatrans->onReturn($this->order, $request);

    $payments = $this->entityTypeManager->getStorage('commerce_payment')->loadMultiple(NULL);
    $this->assertCount(1, $payments);

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = reset($payments);

    $this->assertEquals('authorization', $payment->getState()->value);
    $this->assertEquals(123, $payment->getRemoteId());
    $this->assertEquals($this->order->getTotalPrice(), $payment->getAmount());
    $this->assertEquals('datatrans_id', $payment->getPaymentGatewayId());
  }

  /**
   * @covers ::refundPayment
   */
  public function testRefundPayment() {

    $http_client = $this->prophesize(ClientInterface::class);
    $expected_options = [
      'allow_redirects' => FALSE,
      'auth' => ['', 'secret'],
      'json' => ['amount' => 99900, 'currency' => 'USD', 'refno' => '1'],
      'headers' => ['Content-Type' => 'application/json'],
    ];
    $http_client->request('POST', 'https://api.sandbox.datatrans.com/v1/transactions/123/credit', $expected_options)->willReturn(new Response());
    $this->container->set('http_client', $http_client->reveal());

    $datatrans = $this->getDatatransGateway();

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = Payment::create([
      'state' => 'completed',
      'amount' => $this->order->getTotalPrice(),
      'payment_gateway' => $this->paymentGateway->id(),
      'order_id' => $this->order->id(),
      'test' => TRUE,
      'remote_id' => '123',
      'remote_state' => 'test',
      'authorized' => \Drupal::time()->getRequestTime(),
    ]);
    $payment->save();

    $datatrans->refundPayment($payment);

    $this->assertEquals('refunded', $payment->getState()->value);
    $this->assertEquals(123, $payment->getRemoteId());
    $this->assertEquals($this->order->getTotalPrice(), $payment->getAmount());
    $this->assertEquals($this->order->getTotalPrice(), $payment->getRefundedAmount());
    $this->assertEquals(new Price(0, 'USD'), $payment->getBalance());
    $this->assertEquals('datatrans_id', $payment->getPaymentGatewayId());
  }

  /**
   * @covers ::refundPayment
   */
  public function testRefundPaymentPartial() {

    $http_client = $this->prophesize(ClientInterface::class);
    $expected_options = [
      'allow_redirects' => FALSE,
      'auth' => ['', 'secret'],
      'json' => ['amount' => 12350, 'currency' => 'USD', 'refno' => '1'],
      'headers' => ['Content-Type' => 'application/json'],
    ];
    $http_client->request('POST', 'https://api.sandbox.datatrans.com/v1/transactions/123/credit', $expected_options)->willReturn(new Response());
    $this->container->set('http_client', $http_client->reveal());

    $datatrans = $this->getDatatransGateway();

    /** @var \Drupal\commerce_payment\Entity\PaymentInterface $payment */
    $payment = Payment::create([
      'state' => 'completed',
      'amount' => $this->order->getTotalPrice(),
      'payment_gateway' => $this->paymentGateway->id(),
      'order_id' => $this->order->id(),
      'test' => TRUE,
      'remote_id' => '123',
      'remote_state' => 'test',
      'authorized' => \Drupal::time()->getRequestTime(),
    ]);
    $payment->save();

    $refund_amount = new Price('123.50', 'USD');
    $datatrans->refundPayment($payment, $refund_amount);

    $this->assertEquals('partially_refunded', $payment->getState()->value);
    $this->assertEquals(123, $payment->getRemoteId());
    $this->assertEquals($this->order->getTotalPrice(), $payment->getAmount());
    $this->assertEquals($refund_amount, $payment->getRefundedAmount());
    $this->assertEquals(new Price('875.5', 'USD'), $payment->getBalance());
    $this->assertEquals('datatrans_id', $payment->getPaymentGatewayId());
  }

  /**
   * Returns the datatrans gateway.
   *
   * @return \Drupal\commerce_datatrans\Plugin\Commerce\PaymentGateway\Datatrans
   *   The datatrans gateway.
   */
  protected function getDatatransGateway() {
    $configuration = [
      '_entity' => $this->paymentGateway,
      'password' => 'secret',
    ];
    $gateway_plugin_manager = $this->container->get('plugin.manager.commerce_payment_gateway');
    return $gateway_plugin_manager->createInstance('datatrans', $configuration);
  }

}
